import re
import numpy as np

from typing import Literal

from lucid._tensor import Tensor
from lucid._backend import create_ufunc_op, _FuncOpReturnType
from lucid.types import _EinopsPattern, _NumPyArray

_ReduceStr = Literal["sum", "mean"]


def _parse_pattern(pattern_side: str) -> list[str | tuple[str, ...]]:
    parts = re.findall(r"\([^)]+\)|\S+", pattern_side)
    tokens: list[str | tuple[str, ...]] = []

    for part in parts:
        if part.startswith("(") and part.endswith(")"):
            inner = part[1:-1].strip()
            if not inner:
                raise ValueError("Empty group in pattern.")
            tokens.append(tuple(inner.split()))
        else:
            tokens.append(part)

    return tokens


def _build_intermediate(
    input_tokens: list[str | tuple[str, ...]], shape: tuple[int, ...], shapes: dict
) -> tuple[list[str], list[int]]:
    inter_tokens: list[str] = []
    inter_shape: list[int] = []

    for token, dim in zip(input_tokens, shape):
        if isinstance(token, tuple):
            prod = 1
            group_sizes = []
            for t in token:
                if t not in shapes:
                    raise ValueError(
                        f"Size for token '{t}' in group {token} must be provided."
                    )
                s = shapes[t]
                group_sizes.append(s)
                prod *= s

            if prod != dim:
                raise ValueError(
                    f"Product of sizes {prod} for grouped tokens "
                    + f"{token} does not match merged axis size {dim}."
                )
            inter_tokens.extend(token)
            inter_shape.extend(group_sizes)

        else:
            if token in shapes and shapes[token] != dim:
                raise ValueError(
                    f"Provided size for token '{token}' ({shapes[token]}) "
                    + f"does not match tensor dimension ({dim})."
                )
            inter_tokens.append(token)
            inter_shape.append(dim)

    return inter_tokens, inter_shape


@create_ufunc_op()
def rearrange(
    self: Tensor, pattern: _EinopsPattern, **shapes: int
) -> _FuncOpReturnType:
    try:
        in_pat, out_pat = map(str.strip, pattern.split("->"))
    except Exception as e:
        raise ValueError(
            "Pattern must contain '->' separating input and output patterns."
        ) from e

    input_tokens = _parse_pattern(in_pat)
    output_tokens = _parse_pattern(out_pat)

    if len(input_tokens) != len(self.shape):
        raise ValueError(
            f"Input pattern has {len(input_tokens)} tokens, "
            + f"but tensor has {len(self.shape)} dimensions."
        )

    inter_tokens, inter_shape = _build_intermediate(input_tokens, self.shape, shapes)

    group_splits: list[tuple[int, ...]] = []
    perm: list[int] = []
    used = set()

    for token in output_tokens:
        if isinstance(token, tuple):
            group_perm, group_dims = [], []
            for t in token:
                found = None
                for i, it in enumerate(inter_tokens):
                    if it == t and i not in used:
                        found = i
                        break
                if found is None:
                    raise ValueError(
                        f"Token '{t}' in output group {token} not found in input."
                    )
                group_perm.append(found)
                group_dims.append(inter_shape[found])
                used.add(found)

            group_splits.append(tuple(group_dims))
            perm.extend(group_perm)

        else:
            found = None
            for i, it in enumerate(inter_tokens):
                if it == token and i not in used:
                    found = i
                    break
            if found is None:
                raise ValueError(
                    f"Token '{token}' from output pattern not found in input."
                )

            perm.append(found)
            group_splits.append((inter_shape[found],))
            used.add(found)

    final_shape = tuple(np.prod(group) for group in group_splits)
    intermediate = self.data.reshape(tuple(inter_shape))
    transposed = np.transpose(intermediate, axes=perm)

    result_data = transposed.reshape(final_shape)
    result = Tensor(result_data)

    def compute_grad() -> _NumPyArray:
        unmerged_shape = tuple(dim for group in group_splits for dim in group)
        grad_unmerged = result.grad.reshape(unmerged_shape)

        inv_perm = np.argsort(perm)
        grad_intermediate = np.transpose(grad_unmerged, axes=inv_perm)

        return grad_intermediate.reshape(self.shape)

    return result, compute_grad


@create_ufunc_op()
def reduce(
    self: Tensor, pattern: _EinopsPattern, reduction: _ReduceStr = "sum", **shapes: int
) -> _FuncOpReturnType:
    try:
        in_pat, out_pat = map(str.strip, pattern.split("->"))
    except Exception as e:
        raise ValueError(
            "Pattern must contain '->' separating input and output patterns."
        ) from e

    input_tokens = _parse_pattern(in_pat)
    output_tokens = _parse_pattern(out_pat)

    if len(input_tokens) != len(self.shape):
        raise ValueError(
            f"Input pattern has {len(input_tokens)} tokens, "
            + f"but tensor has {len(self.shape)} dimensions."
        )

    inter_tokens, inter_shape = _build_intermediate(input_tokens, self.shape, shapes)
    kept_indices: list[int] = []
    kept_groups: list[tuple[int, ...]] = []
    used = set()

    for token in output_tokens:
        if isinstance(token, tuple):
            group_inds, group_dims = [], []
            for t in token:
                found = None
                for i, it in enumerate(inter_tokens):
                    if it == t and i not in used:
                        found = i
                        break
                if found is None:
                    raise ValueError(
                        f"Token '{t}' in output group {token} not found in input."
                    )

                group_inds.append(found)
                group_dims.append(inter_shape[found])
                used.add(found)

            kept_indices.extend(group_inds)
            kept_groups.append(tuple(group_dims))

        else:
            found = None
            for i, it in enumerate(inter_tokens):
                if it == token and i not in used:
                    found = i
                    break
            if found is None:
                raise ValueError(
                    f"Token '{token}' from output pattern not found in input."
                )

            kept_indices.append(found)
            kept_groups.append((inter_shape[found],))
            used.add(found)

    all_indices = set(range(len(inter_tokens)))
    reduced_indices = sorted(list(all_indices - used))

    perm = kept_indices + reduced_indices

    kept_flat_shape = tuple(inter_shape[i] for i in kept_indices)
    reduced_shape = tuple(inter_shape[i] for i in reduced_indices)
    final_shape = tuple(np.prod(group) for group in kept_groups)

    intermediate = self.data.reshape(tuple(inter_shape))
    transposed = np.transpose(intermediate, axes=perm)

    reduce_axes = tuple(
        range(len(kept_flat_shape), len(kept_flat_shape) + len(reduced_shape))
    )

    if reduction == "sum":
        reduced_data = np.sum(transposed, axis=reduce_axes)
    elif reduction == "mean":
        reduced_data = np.mean(transposed, axis=reduce_axes)
    else:
        raise ValueError(f"Unsupported reduction method: {reduction}")

    result_data = reduced_data.reshape(final_shape)
    result = Tensor(result_data)

    def compute_grad() -> _NumPyArray:
        grad_kept = result.grad.reshape(kept_flat_shape)
        new_shape = kept_flat_shape + (1,) * len(reduced_shape)

        grad_expanded = grad_kept.reshape(new_shape)
        grad_broadcast = np.broadcast_to(
            grad_expanded, kept_flat_shape + tuple(reduced_shape)
        )

        if reduction == "mean":
            factor = np.prod(reduced_shape) if reduced_shape else 1
            grad_broadcast = grad_broadcast / factor

        inv_perm = np.argsort(perm)
        grad_intermediate = np.transpose(grad_broadcast, axes=inv_perm)

        return grad_intermediate.reshape(self.shape)

    return result, compute_grad


@create_ufunc_op()
def repeat(self: Tensor, pattern: _EinopsPattern, **shapes: int) -> _FuncOpReturnType:
    try:
        in_pat, out_pat = map(str.strip, pattern.split("->"))
    except Exception as e:
        raise ValueError(
            "Pattern must contain '->' separating input and output patterns."
        ) from e

    input_tokens = _parse_pattern(in_pat)
    output_tokens = _parse_pattern(out_pat)

    if len(input_tokens) != len(self.shape):
        raise ValueError(
            f"Input pattern has {len(input_tokens)} tokens, "
            + f"but tensor has {len(self.shape)} dimensions."
        )

    intermediate_tokens, intermediate_shape = _build_intermediate(
        input_tokens, self.shape, shapes
    )

    used_indices = set()
    group_splits: list[tuple[int, ...]] = []
    perm: list[int] = []

    for token in output_tokens:
        if isinstance(token, tuple):
            group_perm, group_dims = [], []
            for t in token:
                found = None
                for i, it in enumerate(intermediate_tokens):
                    if it == t and i not in used_indices:
                        found = i
                        break

                if found is None:
                    group_perm.append(-1)
                    group_dims.append(1)
                else:
                    group_perm.append(found)
                    group_dims.append(intermediate_shape[found])
                    used_indices.add(found)

            group_splits.append(tuple(group_dims))
            perm.extend(group_perm)

        else:
            found = None
            for i, it in enumerate(intermediate_tokens):
                if it == token and i not in used_indices:
                    found = i
                    break

            if found is None:
                perm.append(-1)
                group_splits.append((1,))
            else:
                perm.append(found)
                group_splits.append((intermediate_shape[found],))
                used_indices.add(found)

    base_shape = tuple(dim for group in group_splits for dim in group)
    out_shape_list = []
    idx = 0

    for token in output_tokens:
        if isinstance(token, tuple):
            n = len(token)
            group_perm = perm[idx : idx + n]

            if all(p == -1 for p in group_perm):
                prod_val = 1
                for t in token:
                    if t not in shapes:
                        raise ValueError(
                            f"Size for expansion token '{t}' must be provided."
                        )
                    prod_val *= shapes[t]
                out_shape_list.append(prod_val)

            else:
                prod_val = np.prod(base_shape[idx : idx + n])
                out_shape_list.append(prod_val)

            idx += n
        else:
            if perm[idx] == -1:
                out_shape_list.append(shapes[token])
            else:
                out_shape_list.append(base_shape[idx])
            idx += 1

    out_shape = tuple(out_shape_list)
    kept_order = [p for p in perm if p != -1]

    intermediate = self.data.reshape(tuple(intermediate_shape))
    transposed = (
        np.transpose(intermediate, axes=kept_order) if kept_order else intermediate
    )
    base = transposed.reshape(base_shape)
    tile_multiples = tuple(o // b for b, o in zip(base_shape, out_shape))

    result_data = np.tile(base, tile_multiples)
    result = Tensor(result_data)

    def compute_grad() -> _NumPyArray:
        grad = result.grad
        for i, (b, o) in enumerate(zip(base_shape, out_shape)):
            if b == 1 and o != 1:
                grad = np.sum(grad, axis=i, keepdims=True)

        grad = grad.reshape(base_shape)
        return grad.reshape(self.shape)

    return result, compute_grad



import numpy as np

from lucid._tensor import Tensor
from lucid.types import _ArrayOrScalar


__all__ = ["Parameter", "Buffer"]


class Parameter(Tensor):
    def __init__(self, data: Tensor | _ArrayOrScalar, dtype=np.float32) -> None:
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(data, requires_grad=True, keep_grad=True, dtype=dtype)


class Buffer(Tensor):
    def __init__(self, data: Tensor | _ArrayOrScalar, dtype=np.float32) -> None:
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(data, requires_grad=False, keep_grad=False, dtype=dtype)



from typing import Literal, Sequence
from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _ArrayLikeInt

from lucid._util import func


def reshape(a: Tensor, shape: _ShapeLike) -> Tensor:
    return func._reshape(a, shape)


def squeeze(a: Tensor, axis: _ShapeLike | None = None) -> Tensor:
    return func.squeeze(a, axis)


def unsqueeze(a: Tensor, axis: _ShapeLike) -> Tensor:
    return func.unsqueeze(a, axis)


def ravel(a: Tensor) -> Tensor:
    return func.ravel(a)


def stack(arr: tuple[Tensor, ...], axis: int = 0) -> Tensor:
    return func.stack(*arr, axis=axis)


def hstack(arr: tuple[Tensor, ...]) -> Tensor:
    return func.hstack(*arr)


def vstack(arr: tuple[Tensor, ...]) -> Tensor:
    return func.vstack(*arr)


def concatenate(arr: tuple[Tensor, ...], axis: int = 0) -> Tensor:
    return func.concatenate(*arr, axis=axis)


def pad(a: Tensor, pad_width: _ArrayLikeInt) -> Tensor:
    return func.pad(a, pad_width)


def repeat(a: Tensor, repeats: int | Sequence[int], axis: int | None = None) -> Tensor:
    return func.repeat(a, repeats, axis=axis)


def tile(a: Tensor, reps: int | Sequence[int]) -> Tensor:
    return func.tile(a, reps)


def flatten(a: Tensor) -> Tensor:
    return func.flatten(a)


def meshgrid(
    a: Tensor, b: Tensor, indexing: Literal["xy", "ij"] = "xy"
) -> tuple[Tensor, Tensor]:
    return func.meshgrid(a, b, indexing)


def split(
    a: Tensor, size_or_sections: int | list[int] | tuple[int], axis: int = 0
) -> tuple[Tensor, ...]:
    return func.split(a, size_or_sections, axis)


def tril(a: Tensor, diagonal: int = 0) -> Tensor:
    return func.tril(a, diagonal)


def triu(a: Tensor, diagonal: int = 0) -> Tensor:
    return func.triu(a, diagonal)


def broadcast_to(a: Tensor, shape: _ShapeLike) -> Tensor:
    return func.broadcast_to(a, shape)


Tensor.reshape = func._reshape_inplace
Tensor.squeeze = func.squeeze
Tensor.unsqueeze = func.unsqueeze
Tensor.ravel = func.ravel
Tensor.pad = func.pad
Tensor.repeat = func.repeat
Tensor.tile = func.tile
Tensor.flatten = func.flatten
Tensor.split = func.split
Tensor.tril = func.tril
Tensor.triu = func.triu
Tensor.broadcast_to = func.broadcast_to



from typing import Any, Callable, Dict, Sequence
import numpy as np

_base_dtype: np.floating = np.float32

_Scalar = int | float
_NumPyArray = np.ndarray
_ArrayOrScalar = _Scalar | list[_Scalar] | _NumPyArray

_ShapeLike = list[int] | tuple[int]

_ArrayLike = list | _NumPyArray
_ArrayLikeInt = int | Sequence[int | tuple[int, int]]

_StateDict = Dict[str, Any]

_OptimClosure = Callable[[], Any]

_EinopsPattern = str


def _change_base_dtype(dtype: type) -> None:
    _base_dtype = dtype



from typing import Literal

from lucid._tensor import Tensor
from lucid.einops import _func

from lucid.types import _EinopsPattern

_ReduceStr = Literal["sum", "mean"]


def rearrange(a: Tensor, pattern: _EinopsPattern, **shapes: int) -> Tensor:
    return _func.rearrange(a, pattern, **shapes)


def reduce(
    a: Tensor, pattern: _EinopsPattern, reduction: _ReduceStr = "sum", **shapes: int
) -> Tensor:
    return _func.reduce(a, pattern, reduction, **shapes)


def repeat(a: Tensor, pattern: _EinopsPattern, **shapes: int) -> Tensor:
    return _func.repeat(a, pattern, **shapes)



"""
# `Lucid `

**Lucid** is an educational deep learning framework developed to help users understand
the underlying mechanics of deep learning models and tensor operations.

It is designed to provide a simple yet powerful environment to experiment with neural networks,
optimization, and backpropagation using only `NumPy`.

Lucid is ideal for those who want to learn about the inner workings of deep learning
algorithms and operations without the complexity of high-level frameworks.

[📑 Lucid Documentation](https://chanlumerico.github.io/lucid/build/html/index.html)
"""

from contextlib import contextmanager
from typing import Any, Generator, SupportsIndex, Callable
from functools import wraps
from pathlib import Path

import os
import sys
import json
import numpy as np

from lucid._tensor import Tensor
from lucid._func import *
from lucid._util import *

from lucid.types import _ArrayOrScalar, _NumPyArray, _ArrayLike, _ShapeLike

import lucid.linalg as linalg
import lucid.random as random
import lucid.einops as einops
import lucid.nn as nn

_grad_enabled: bool = True

newaxis = np.newaxis

pi = np.pi
inf = np.inf


def tensor(
    data: Tensor | _ArrayOrScalar,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: Any = np.float32,
) -> Tensor:
    if isinstance(data, Tensor):
        data = data.data
    return Tensor(data, requires_grad, keep_grad, dtype)


def to_tensor(
    a: _ArrayLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: Any = np.float32,
) -> Tensor:
    return tensor(a, requires_grad, keep_grad, dtype)


@contextmanager
def no_grad() -> Generator:
    global _grad_enabled
    prev_state = _grad_enabled
    _grad_enabled = False
    try:
        yield
    finally:
        _grad_enabled = prev_state


def grad_enabled() -> bool:
    return _grad_enabled


def shape(a: Tensor | _NumPyArray) -> _ShapeLike:
    if hasattr(a, "shape"):
        return a.shape

    raise ValueError(f"The argument must be a Tensor or a NumPy array.")


def _check_input_dim(tensor: Tensor, dim: int) -> None:
    if tensor.ndim != dim:
        raise ValueError(f"expected {dim}D input (got {tensor.ndim}D input).")


def _set_tensor_grad(
    tensor: Tensor, grad: _NumPyArray, at: SupportsIndex = ...
) -> None:
    if tensor.requires_grad:
        if tensor.grad is None:
            tensor.grad = grad
        else:
            if not tensor.grad.flags.writeable:
                tensor.grad = tensor.grad.copy()
            tensor.grad[at] = tensor.grad[at] + grad


def _check_is_tensor(any: Tensor | _ArrayOrScalar) -> Tensor:
    if not isinstance(any, Tensor):
        return Tensor(any)
    return any


def _match_grad_shape(data: _NumPyArray, grad: _NumPyArray) -> _NumPyArray:
    target = data.shape
    nd_target = len(target)

    if grad.ndim < nd_target:
        grad = grad.reshape((1,) * (nd_target - grad.ndim) + grad.shape)

    elif grad.ndim > nd_target:
        extra = grad.ndim - nd_target
        extra_dims = grad.shape[:extra]
        if any(dim != 1 for dim in extra_dims):
            grad = grad.sum(axis=tuple(range(extra)))
        else:
            grad = grad.reshape(grad.shape[extra:])

    for i in range(nd_target):
        g_dim = grad.shape[i]
        t_dim = target[i]
        if g_dim == t_dim:
            continue
        elif g_dim > t_dim:
            if t_dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
            else:
                if g_dim % t_dim == 0:
                    grad = grad.sum(axis=i, keepdims=True)
                else:
                    raise ValueError(
                        f"Cannot match grad shape on axis {i}: "
                        + f"grad has {g_dim} while data has {t_dim}"
                    )

        elif g_dim < t_dim:
            if g_dim == 1:
                continue
            else:
                raise ValueError(
                    f"Cannot match grad shape on axis {i}: "
                    + f"grad has {g_dim} while data has {t_dim}"
                )

    grad = np.broadcast_to(grad, target)
    return grad


def _get_overloaded_shape(args: int | _ShapeLike) -> _ShapeLike:
    if len(args) == 1 and isinstance(args[0], (tuple, list)):
        shape = tuple(args[0])
    else:
        shape = tuple(args)
    return shape


REGISTRY_PATH: Path = Path("lucid/models/registry.json")

_ModuleReturnFunc = Callable[[Any], nn.Module]


def register_model(func: _ModuleReturnFunc) -> _ModuleReturnFunc:
    @wraps(func)
    def wrapper(*args, **kwargs) -> nn.Module:
        if os.environ.get("SPHINX_BUILD"):
            return func(*args, **kwargs)

        if not REGISTRY_PATH.exists():
            REGISTRY_PATH.parent.mkdir(parents=True, exist_ok=True)
            with open(REGISTRY_PATH, "w") as f:
                json.dump({}, f)

        with open(REGISTRY_PATH, "r") as f:
            registry = json.load(f)

        model = func(*args, **kwargs)
        model._alt_name = func.__name__
        name = func.__name__

        if name in registry:
            return model

        family = model.__class__.__name__
        param_size = model.parameter_size
        arch = sys.modules[func.__module__].__package__.replace("lucid.models.", "")

        registry[name] = dict(
            name=name,
            family=family,
            param_size=param_size,
            arch=arch,
        )

        with open(REGISTRY_PATH, "w") as f:
            json.dump(registry, f, indent=4)

        return model

    return wrapper
